
/*
 * Date (yymmdd):               100208
 * Written by:                  Dr. Francisco J. Romero-Campero
 * Additional contributions by: Jonathan Blakes
 * (in no particular order)     Dr. Hongqing Cao
 *                              Dr. Natalio Krasnogor
 *				Dr. Jamie Twycross
 * Funded by:                   EPSRC grant EP/E017215/1
 * Website:                     http://www.infobiotic.org/
 * Contact:                     nxk@cs.nott.ac.uk
 */

#if HAVE_CONFIG_H
        #include <config.h>
#endif

#include <fstream>
#include <pmodelchecker/prismModelChecker.h>
#include <pmodelchecker/prismmodel.h>
#include <pmodelchecker/formulaBNF.h>


using namespace PMODELCHECKER;

void PMODELCHECKER::usePRISM(MCParameters* MCParams)
{
    //Independently of the task to carry we translate our P system specifications
    //into a PRISM model.
    ECSB::LPPsystem Pmodel(MCParams->model_specification);
    PrismModel prismModel(Pmodel);

    //PRISM command to perform the specific task
    std::string command;

    //Output of the system call
    int output;

    //prismModel.show();
    prismModel.write(MCParams->PRISM_model);

    //XML file containing the variables of the model for communication with
    //property builder tool
    prismModel.writeParameters(MCParams->parameters_file);

    if (MCParams->task == "Build")
    {
        //Build the model by system calling prism PRISM_model parameters export_options
        command = "prism ";
        command += MCParams->PRISM_model;
        command += " -const ";
        command += MCParams->model_parameters;
        command += ",";
        command += MCParams->formula_parameters;
        command += " -exportstates ";
        command += MCParams->states_file;
        command += " -exporttrans ";
        command += MCParams->transitions_file;

        output = system(command.c_str());
    }
    else if (MCParams->task=="Verify")
    {
        //Generate the temporal formulas file for PRISM
        std::string parameters = generateFormulas(MCParams->temporal_formulas);

        //Verify properties specified in a file by system calling
        //prism model_file property_file parameters
        command = "prism ";
        command += MCParams->PRISM_model;
        command += " ";
        command += MCParams->temporal_formulas;
        command += " -const ";
        command += MCParams->model_parameters;
        command += ",";
//        command += MCParams->formula_parameters;
        command += parameters;
        command += " -exportstates ";
        command += MCParams->states_file;
        command += " -exporttrans ";
        command += MCParams->transitions_file;
        command += " -exportresults ";
        command += MCParams->results_file;

        output = system(command.c_str());
        
        //Remove temporal file with the formulas

        if( remove("formulas.temp") != 0 )
        {
            cerr << "Unable to remove temporary file: formulas.temp" << std::endl;
        }

    }
    else if (MCParams->task == "Approximate")
    {
        //Generate the temporal formulas file for PRISM
        std::string parameters = generateFormulas(MCParams->temporal_formulas);

        //Approximate properties specified in a file by system calling
        //prism model_file property_file -sim -simsamples
        command = "prism ";
        command += MCParams->PRISM_model;
        command += " ";
        command += "formulas.temp";
        command += " -const ";
        command += MCParams->model_parameters;
        command += ",";
//        command += MCParams->formula_parameters;
        command += parameters;
        command += " -sim ";
        command += " -simsamples ";
        command += somethingToString<unsigned long>(MCParams->number_samples);
        command += " -simconf ";
        command += somethingToString<float>(MCParams->confidence);
        command += " -simpathlen 10000000";
        command += " -exportresults ";
        command += MCParams->results_file;

        output = system(command.c_str());

        //Remove temporal file with the formulas

        if( remove("formulas.temp") != 0 )
        {
            cerr << "Unable to remove temporary file: formulas.temp" << std::endl;
        }
    }

}


std::string PMODELCHECKER::generateFormulas(std::string temporalFormulas)
{
    //Get content of file
    std::string fileContent = fileToString(temporalFormulas);

    //Parse the file
    formulas_p formulasParser;
    tree_parse_info<> result = boost::spirit::classic::ast_parse(fileContent.c_str(),formulasParser,space_p);

    //Checking if parsing has been successful
    if(!result.full)
    {
        cerr << "Error parsing file " << temporalFormulas << " at:" << std::endl;
        std::cout << result.stop << std::endl;
        exit(EXIT_FAILURE);
    }

    //The abstract syntax tree associated with the parsing of the formulas
    iter_t astTree = result.trees.begin();

    //Container for formula and parameters
    std::map<std::string,std::map<std::string, std::vector<std::string> > >
            formulas;

    for(iter_t formula = astTree->children.begin();
        formula != astTree->children.end();
        ++formula)
    {
        std::string formulaSpec(formula->value.begin(), formula->value.end());
        replaceOperators(&formulaSpec);

        //Parameters associated with the formula
        std::map< std::string, std::vector<string> > formulaParameters;

        iter_t parameters = formula->children.begin();

        //For each parameter we collect its name, lower and upper bound and incremental step
        for(iter_t param = parameters->children.begin(); param != parameters->children.end(); ++param )
        {
            std::string parameterName(param->value.begin(),param->value.end());

            std::vector<std::string> parameterValues;

            iter_t lowerBoundDef = param->children.begin();
            std::string lowerBound(lowerBoundDef->value.begin(),lowerBoundDef->value.end());
            parameterValues.push_back(lowerBound);

            iter_t stepDef = param->children.begin() + 1;
            std::string step(stepDef->value.begin(),stepDef->value.end());
            parameterValues.push_back(step);

            iter_t upperBoundDef = param->children.begin() + 2;
            std::string upperBound(upperBoundDef->value.begin(),upperBoundDef->value.end());
            parameterValues.push_back(upperBound);

            //Inserting the parameter
            formulaParameters[parameterName] = parameterValues;
        }

        formulas[formulaSpec] = formulaParameters;

    }

    //Write parameters in file and generate string with constant ranges
    ofstream temporalFile("formulas.temp",ios::out);
    std::string constantRanges;


		std::map<std::string,std::vector<std::string> > simplifiedParameters = removeParamDuplicates(formulas);

        for(std::map<std::string, std::vector<std::string> >::iterator param = simplifiedParameters.begin();
            param != simplifiedParameters.end();
            ++param)
        {
            temporalFile << "const int " << param->first << ";" << std::endl;
            constantRanges += param->first + "=";
            constantRanges += param->second[0] + ":" + param->second[1] + ":" + param->second[2] + ",";
        }

    //Write the formulas
    temporalFile << std::endl;
    for(std::map<std::string,std::map<std::string,std::vector<std::string> > >::iterator formula = formulas.begin();
        formula != formulas.end(); ++formula)
    {

        temporalFile << formula->first << std::endl;
    }

    temporalFile.close();

    return constantRanges;
}

std::map<std::string,std::vector<std::string> > PMODELCHECKER::removeParamDuplicates(std::map<std::string,std::map<std::string,std::vector<std::string> > > formulas)
{

	std::map<std::string,std::vector<std::string> > simplifiedParameters;
   	
   for(std::map<std::string,std::map<std::string,std::vector<std::string> > >::iterator formula = formulas.begin();
        formula != formulas.end(); ++formula)
    {
		for(std::map<std::string, std::vector<std::string> >::iterator param = formula->second.begin();
            param != formula->second.end();
            ++param)
        {

			std::map<std::string,std::vector<std::string> >::iterator paramIt = simplifiedParameters.find(param->first);
			if(paramIt == simplifiedParameters.end())
			{
				simplifiedParameters[param->first] = param->second;
			}
			else if(param->second != simplifiedParameters[param->first])
			{
				std::vector<std::string> values = simplifiedParameters[param->first];

				cerr << " Error: parameters with same name and different ranges: " << std::endl;
				std::cout << param->second[0] << ":" << param->second[1] << ":" << param->second[2];
				std::cout << values[0] << ":" << values[1] << ":" << values[2];

		        exit(EXIT_FAILURE);
			}

        }
    }

	return simplifiedParameters;
}


void PMODELCHECKER::replaceOperators(std::string* formula)
{
    //Replace and symbol ^ for &
    size_t pos = formula->find("^");
    while( pos != formula->npos)
    {
        formula->replace(pos,1,"&");
        pos = formula->find("^",pos);
    }
}
